#include <stdio.h>
#include <string.h>
#include <time.h>

#include "plfunc.h"

Object reduce(Object* params, unused int length, struct Environment* env)
{
    checkTypes(reduce)
    Object list = params[0];
    Object func = params[1];
    Object total = params[2];

    if (list.type != TYPE_LIST) {
        func = cloneObject(func);
        list.forward = &total;
        return funcyEval(&func, &list, 2, env);
    }

    Object* first = list.list;
    FOR_POINTER_IN_LIST(&list) {
        func = cloneObject(func);
        Object oldTotal = total;
        total.forward = POINTER;
        total = funcyEval(&func, &total, 2, env);
        if (POINTER != first) {
            cleanObject(&oldTotal);
        }
    }

    return total;
}

Object charAt(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(charAt)
    Object string = params[0];
    Object at = params[1];

    Object c = withLen(1, string.type);
    c.string[1] = '\0';

    for (int i = 0; i < at.number; i++) {
        if (string.string[i] == '\0') {
            c.string[0] = '\0';
            return c;
        }
    }

    c.string[0] = string.string[at.number];
    return c;
}

Object filter(Object* params, unused int length, struct Environment* env)
{
    checkTypes(filter)
    Object condition = params[0];
    Object list = params[1];

    BuildListNamed(filtered);
    FOR_POINTER_IN_LIST(&list) {
        Object cloned = cloneObject(condition);
        Object first_eval = eval(&cloned, env);
        Object testee = cloneObject(*POINTER);
        Object e = funcyEval(&first_eval, &testee, 1, env);
        if (e.number) {
            addToList(filtered, testee); // May need to re-clone testee?
        }
        cleanObject(&e);
        cleanObject(&cloned);
    }
    return filtered;
}

Object append(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(append)
    Object list = params[0];
    Object newElement = params[1];

    Object newList = cloneObject(list);
    nf_addToList(&newList, cloneObject(newElement));
    return newList;
}

Object prepend(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(prepend)
    Object list = params[0];
    Object newElement = params[1];

    Object newList = listObject();
    nf_addToList(&newList, cloneObject(newElement));
    appendList(&newList, &list);
    return newList;
}

Object at(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(at)
    Object index = params[0];
    Object list = params[1];

    const Object* found = itemAt(&list, index.number);

    if (found) {
        return cloneObject(*found);
    }

    return errorObject(INDEX_PAST_END);
}

Object rest(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(rest)
    Object list = params[0];

    BuildListNamed(ret);
    Object* l = &list;
    FOR_POINTER_IN_LIST(l) {
        if (POINTER == l->list) {
            continue;
        }
        addToList(ret, cloneObject(*POINTER));
    }
    return ret;
}

Object reverse(Object* params, unused int length, unused struct Environment* ignore2)
{
    checkTypes(reverse)
    Object _list = params[0];

    const Object* list = &_list;
    Object rev = listObject();

    Object* tail = NULL;
    FOR_POINTER_IN_LIST(list) {
        Object* oldTail = tail;
        allocObject(&tail, cloneObject(*POINTER));
        if (oldTail) {
            tail->forward = oldTail;
        }
    }

    rev.list = tail;
    return rev;
}

Object isNum(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(isNum)
    Object test = params[0];

    return boolObject(test.type == TYPE_NUMBER);
}

Object isList(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(isList)
    Object test = params[0];

    return boolObject(test.type == TYPE_LIST);
}

Object isString(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(isString)
    Object test = params[0];

    return boolObject(test.type == TYPE_STRING);
}

Object isErr(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(isErr)
    Object test = params[0];

    return boolObject(test.type == TYPE_ERROR);
}

Object charVal(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(charVal)
    Object test = params[0];

    return numberObject(test.string[0]);
}

Object chars(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(chars)
    char c[2];
    c[1] = '\0';
    int i = 0;
    Object list = listObject();
    while (params[0].string[i]) {
        c[0] = params[0].string[i];
        nf_addToList(&list, stringFromSlice(c, 1));
        i++;
    }
    return list;
}

size_t literalLen(const char* pattern)
{
    const char* cursor = pattern;
    while (*cursor != '*' && *cursor != '\0') {
        cursor++;
    }
    return cursor - pattern;
}

/// Returns 1 on match, 0 otherwise
int stringComp(const char* string, const char* pattern)
{
    const char* cursor = string;
    const char* patternCursor = pattern;

    while (*patternCursor) {
        size_t len = literalLen(patternCursor);
        if (len == 0) {
            patternCursor += 1;
            while (!stringComp(cursor, patternCursor)) {
                cursor++;
                if (*cursor == '\0') {
                    return 0;
                }
            }
        } else {
            if (strncmp(patternCursor, cursor, len) != 0) {
                return 0;
            }
            patternCursor += len;
        }
    }
    return 1;
}

Object matches(Object* params, int length, unused struct Environment* env)
{
    checkTypes(matches)
    return boolObject(stringComp(params[0].string, params[1].string));
}

Object slen(Object* params, int length, unused struct Environment* env)
{
    checkTypes(slen)
    return numberObject(strlen(params[0].string));
}

Object substring(Object* params, int length, unused struct Environment* env)
{
    checkTypes(substring)
    Object start = params[0]; // First char to include
    Object end = params[1];   // First char to exclude
    Object string = params[2];

    long len = end.number - start.number;
    size_t stringLen = strlen(string.string);
    if (len < 0) {
        throw(BAD_PARAMS, "substr start index (%ld) is higher than its end index (%ld)!", start.number, end.number);
    }
    if (start.number > stringLen) {
        throw(BAD_PARAMS, "substr start index (%ld) is higher than the string length (%lu)", start.number, stringLen);
    }

    if (len > stringLen - start.number) {
        len = stringLen - start.number;
    }

    Object substr = withLen(len, TYPE_STRING);
    snprintf(substr.string, len + 1, "%s", string.string + start.number);
    return substr;
}

Object printEnvO(Object* params, int length, unused struct Environment* env)
{
    int printPointers = 0;
    if (length > 0 && params[0].type == TYPE_BOOL) {
        printPointers = params[0].number;
    }
    printEnv(global(), printPointers);
    return numberObject(0);
}

Object parseEvalO(Object* params, unused int length, struct Environment* env)
{
    Object text = params[0];

    env = global();
    switch (text.type) {
        case TYPE_SYMBOL: {
            Object string = eval(&text, env);
            Object parsed = parseEval(string.string, env);
            cleanObject(&string);
            return parsed;
        }
        case TYPE_SLIST:
            return evalList(&text, env);
        case TYPE_STRING:
            return parseEval(text.string, env);
        default:
            throw(CAN_ONLY_EVAL_STRINGS, "Tried to (eval) a %s, instead of a string or symbol list", getTypeName(&text));
    }
}

Object catObjects(Object* params, int length, unused struct Environment* env)
{
    checkTypes(catObjects)

    if (length == 0) {
        return stringFromSlice("", 0);
    }

    char* strings[length];
    size_t totalLength = 0;
    for (int i = 0; i < length; i++) {
        strings[i] = stringObj(&params[i], &totalLength);
    }

    Object string = withLen(totalLength, TYPE_STRING);
    string.string[0] = '\0';

    for (int i = 0; i < length; i++) {
        strcat(string.string, strings[i]);
        free(strings[i]);
    }

    return string;
}


Object len(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(len)

    return numberObject(listLength(&params[0]));
}

#define BASIC_MATH(_name, _op) \
Object _name(Object* params, int length, struct Environment* env) \
{ \
    if (length == 0) { \
        return numberObject(0); \
    } \
    Object sum = numberObject(params[0].number); \
    for (int i = 1; i < length; i++) { \
        sum.number _op params[i].number; \
    } \
    return sum; \
}

BASIC_MATH(add, +=)

BASIC_MATH(sub, -=)

BASIC_MATH(mul, *=)

BASIC_MATH(dvi, /=)

BASIC_MATH(mod, %=)

int areEqual(const Object* obj1, const Object* obj2);

int listEquality(const Object* list1, const Object* list2)
{
    Object* element1, * element2;
    for (element1 = (list1)->list, element2 = (list2)->list;
         element1 != ((void*) 0) && element2 != ((void*) 0);
         element1 = element1->forward, element2 = element2->forward) {
        if (!areEqual(element1, element2)) {
            return 0;
        }
    }
    return (element1 == NULL && element2 == NULL);
}

int areEqual(const Object* obj1, const Object* obj2)
{
    const long n1 = obj1->number;
    const long n2 = obj2->number;

    if (bothAre(TYPE_STRING, obj1, obj2)) {
        return !strcmp(obj1->string, obj2->string);
    }
    if (bothAre(TYPE_LIST, obj1, obj2)) {
        return listEquality(obj1, obj2);
    }
    return n1 == n2 && areSameType(obj1, obj2);
}

Object equ(Object* params, int length, unused struct Environment* env)
{
    if (length < 2) {
        return errorObject(NOT_ENOUGH_ARGUMENTS);
    }

    for (int i = 0; i < length - 1; i++) {
        if (!areEqual(&params[i], &params[i + 1])) {
            return falseObject();
        }
    }

    return trueObject();
}

Object or(Object* params, int length, unused struct Environment* env)
{
    if (length < 2) {
        return errorObject(NOT_ENOUGH_ARGUMENTS);
    }

    for (int i = 0; i < length - 1; i++) {
        if (params[i].number || params[i + 1].number) {
            return trueObject();
        }
    }
    return falseObject();
}

#define BASIC_COMPARISON(NAME, OP)\
Object NAME(Object* params, int length, struct Environment* env) \
{ \
    if (length < 2) { \
        return errorObject(NOT_ENOUGH_ARGUMENTS); \
    } \
    \
    for (int i = 0; i < length - 1; i++) {                       \
        if (!(params[i].number OP params[i + 1].number)) { \
            return falseObject(); \
        } \
    } \
    return trueObject(); \
}

BASIC_COMPARISON(greaterThan, >)

BASIC_COMPARISON(lessThan, <)

BASIC_COMPARISON(and, &&)

int timeStructDefinition = -1;

Object getTime(unused Object* params, unused int length, struct Environment* env)
{
    if (timeStructDefinition == -1) {
        parseEval("(struct Time (minute hour sec))", env);
        timeStructDefinition = getStructIndex("Time");
    }
    time_t t = time(NULL);
    struct tm tm = *localtime(&t);
    Object o = structObject(timeStructDefinition);
    o.structObject->fields[0] = numberObject(tm.tm_min);
    o.structObject->fields[1] = numberObject(tm.tm_hour);
    o.structObject->fields[2] = numberObject(tm.tm_sec);
    return o;
}

#ifdef STANDALONE

#include <unistd.h>

Object print(Object* params, int length, unused struct Environment* env)
{
    for (int i = 0; i < length; i++) {
        _printObj(&params[i], 0);
    }
    return numberObject(0);
}

Object numToChar(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(numToChar)
    Object c = params[0];

    if (c.number > 255 || c.number < 0) {
        throw(BAD_NUMBER, "Char values should be between 0 and 255 (inclusive), but received %ld", c.number);
    }
    char ch[1] = { c.number };
    return stringFromSlice(ch, 1);
}

Object takeInput(Object* params, int length, unused struct Environment* env)
{
    Object prompt = params[0];
    if (length > 0 && prompt.type == TYPE_STRING) {
        printf("%s", prompt.string);
    }
    char input[256] = "";
    if (fgets(input, 256, stdin)) {
        return stringFromSlice(input, strlen(input) - 1);
    }
    return errorWithContext(NULL_PARSE, "fgets() error");
}

Object cd(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(cd)

    return numberObject(chdir(params[0].string));
}

Object cwd(unused Object* params, unused int length, unused struct Environment* env)
{
    char c[128];
    return nullTerminated(getcwd(c, sizeof(c)));
}

Object systemCall(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(systemCall)
    Object process = params[0];

    if (isStringy(process)) {
        return numberObject(system(process.string));
    }
    return numberObject(255);
}

char* readFileToString(FILE* input)
{
    size_t capacity = 128;
    char* string = malloc(sizeof(char) * capacity);
    int c;
    int i = 1; // Skip refCount

    string[0] = 1; // Set refCount
    while ((c = fgetc(input)) != EOF) {
        string[i] = c;
        i++;
        if (i == capacity) {
            char* prev = string;
            capacity *= 2;
            string = malloc(sizeof(char) * capacity);
            memcpy(string, prev, sizeof(char) * (capacity / 2));
            free(prev);
        }
    }
    string[i] = '\0';

    return string + 1; // Offset past refCount
}

Object readFileToObject(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(readFileToObject)
    Object filename = params[0];

    FILE* file = fopen(filename.string, "r");
    if (!file) {
        throw(NULL_PARSE, "Error opening file at %s", filename.string);
    }

    Object string = newObject(TYPE_STRING);
    string.string = readFileToString(file);
    return string;
}

Object getEnvVar(Object* params, unused int length, unused struct Environment* env)
{
    checkTypes(getEnvVar)
    const char* envVar = getenv(params[0].string);
    if (envVar) {
        return nullTerminated(envVar);
    }
    return stringFromSlice("", 0);
}

#endif // STANDALONE
